#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>

# ------------
#| Change Log |
# ------------
# Rel 1: Initial release /18/11/2021/ https://www.gimpscripts.net/2021/11/pattern-editor.html
#               inspired Script-fu Scale Pattern by Rob Antonishen http://gimpchat.com/viewtopic.php?f=9&t=17348#p238328
# Used codes: - fill_rotate.py https://gimplearn.net/viewtopic.php?f=3&t=1393 by Tin Tran.
#             - fu_preview (modificed gimpfu) by Claude Lion [aka cli345] http://gimpchat.com/viewtopic.php?f=9&t=19420
#               & gimfu3 with tabs  https://www.arakne.es/en/gimp-plugin-follow-path-map-values-on-the-y-coordinate/
#             - gegl_command by Kevin Payne & Claude Lion [cli345]: http://gimpchat.com/viewtopic.php?f=9&t=18040&start=30#p264354 
# Rel 2: /22/11/2021/:
#              - change initial Preview to 512 px
#              - add options SYMETRIZE.
# Rel 3: /25/11/2021/:
#              - Change of order tab "Fill & Save"
#              - add tab "Kaleidoscop"
# Rel 4: /28/11/2021/:
#              - add tab "G'MIC Array [Mirrored]"
# Rel 5: /29/08/2022/: 
#              - preview is disabled by default after running the plug-in
#              - some tabs have been reorganized
# 			   - bug fix "expected tile ack and received: 5" (now only if the active layer has an alpha channel)
#			   - add tab "Color"
# Rel 6: /29/08/2022/: 
#              - bugfix for Rotate (angle) if selected "Pattern",
#			   - simplify the tab 'Symmetrize',
#		       - supplementing the Help tab.
# Rel 7: /31/08/2022/:
#             - order of tab has been changed
#             - effects added: "Alien Map","Camouflage","Canvas","Crystal","Cubism","Denim","Edge","Edge Neon","Edge Sobel","Engrave","Mosaic",
#               "Pixelize","Plasma","Rainbow","Value Invert","Video Degradation" and the ability to adjust their opacity.
#			  - in the 'Symmetrize' tab a bug for Patterns with transparency has been fixed
# Rel 8: /03/09/2022/:
#             - to tab "Effect" add: Blend Mode, Brightness, Contrast.
# Rel 9: /21/09/2022/:
#             - to tab "Effect" add: Checkered:"From Colors Layer" or "From Palette" based on palette-to-textile.py by Giuseppe Conte and palette_from_image.py by Tin Tran
#####################################################################################################################################################
############################################# START fu_preview 0.1.1 by Claude Lion $ J.F. Garcia ###################################################
#####################################################################################################################################################

import string as _string
import math
import gimp
import gimpcolor
import colorsys
import re
from gimpenums import *
pdb = gimp.pdb
import os, time, sys
import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)	#MrQ for translate in dir "locle"
_ = t.ugettext																	#MrQ for translate in dir "locle"
import random
class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}


def register(proc_name, blurb, help, author, copyright, date, label,
             imagetypes, params, results, function,
             menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
                return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements "
                          "(%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead"
                       % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright,
                                       date, label, imagetypes,
                                       plugin_type, params, results,
                                       function, menu, domain,
                                       on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date,
         label, imagetypes, plugin_type,
         params, results, function, menu, domain,
         on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]],
                     x[1],
                     _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode",
                                     "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright,
                               date, label, imagetypes, plugin_type,
                               params, results)

        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date,
     label, imagetypes, plugin_type,
     params, results, function, menu, domain,
     on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
#    import pango
    gimpui.gimp_ui_init () # add MrQ
    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default="", **kwargs): # modif C.L.
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            self.set_activates_default(True)
            #~ self.connect('paste-clipboard', lambda _: kwargs['dialg'].la_previ.invalidate()) # modif C.L.
            self.get_buffer().connect('inserted-text', lambda _1, _2, _3, _4: kwargs['dialg'].la_previ.invalidate()) # modif C.L.
            self.get_buffer().connect('deleted-text', lambda _1, _2, _3: kwargs['dialg'].la_previ.invalidate()) # modif C.L

        def get_value(self):
            return self.get_text()


    class LabelTitle(gtk.Label):
        #def __init__ (self, default="", **kwargs): # modif C.L. ???
        def __init__(self, default="", align = (0.0, 0.0)):		
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default="", **kwargs): # modif C.L.
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))
            self.buffer.connect('changed', lambda _: kwargs['dialg'].la_previ.invalidate()) # modif C.L.			

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5), **kwargs): # modif C.L.
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1],
                                      step, 10 * step, 0)
            # modif C.L.
            def react(adj):
                return self.get_toplevel().la_previ.invalidate()
            #~ self.adj.connect('value-changed', lambda adj: self.get_toplevel().la_previ.invalidate())
            self.adj.connect('value-changed', react)
            # end modif C.L.
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=10, bounds=(0, 100, 5), **kwargs): # modif C.L.
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1],
                                      step, 10 * step, 0)
            self.adj.connect('value-changed', lambda adj: kwargs['dialg'].la_previ.invalidate()) # modif C.L.
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))


    class ToggleEntry(gtk.ToggleButton):
        #~ def __init__(self, default=0):
        def __init__(self, default=0,labels=[_("DON'T USE"),_("YES")], **kwargs): # modif C.L.
            
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()

            #~ self.connect("toggled", self.changed)
            # modif C.L.
            def react(tog):
                self.changed(tog)
                return kwargs['dialg'].la_previ.invalidate()
            self.connect("toggled", react)
            # end modif C.L.


            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("YES"), 1), (_("No"), 0)), **kwargs): # modif C.L.
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=(), **kwargs): # modif C.L.
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)
            self.connect('changed', lambda adj: kwargs['dialg'].la_previ.invalidate()) # modif C.L.

        def get_value(self):
            return self.get_active()

    def FileSelector(default="", title=None, **kwargs): # modif C.L.
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            if default == "/": default = ""
            return DirnameSelector(default)
        else:
            return FilenameSelector(default, title=title, save_mode=False)

    class FilenameSelector(gtk.HBox):
        #gimpfu.FileChooserButton
        def __init__(self, default, save_mode=True, title=None, **kwargs): # modif C.L.
            super(FilenameSelector, self).__init__()
            if not title:
                self.title = _("Python-Fu File Selection")
            else:
                self.title = title
            self.save_mode = save_mode
            box = self
            self.entry = gtk.Entry()
            image = gtk.Image()
            image.set_from_stock(gtk.STOCK_FILE, gtk.ICON_SIZE_BUTTON)
            self.button = gtk.Button()
            self.button.set_image(image)
            box.pack_start(self.entry)
            box.pack_start(self.button, expand=False)
            self.button.connect("clicked", self.pick_file)
            if default:
                self.entry.set_text(default)

        def show(self):
            super(FilenameSelector, self).show()
            self.button.show()
            self.entry.show()

        def pick_file(self, widget):
            entry = self.entry
            dialog = gtk.FileChooserDialog(
                         title=self.title,
                         action=(gtk.FILE_CHOOSER_ACTION_SAVE
                                     if self.save_mode else
                                 gtk.FILE_CHOOSER_ACTION_OPEN),
                         buttons=(gtk.STOCK_CANCEL,
                                gtk.RESPONSE_CANCEL,
                                gtk.STOCK_SAVE
                                     if self.save_mode else
                                gtk.STOCK_OPEN,
                                gtk.RESPONSE_OK)
                        )
            dialog.set_alternative_button_order ((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))
            dialog.show_all()
            response = dialog.run()
            if response == gtk.RESPONSE_OK:
                entry.set_text(dialog.get_filename())
            dialog.destroy()

        def get_value(self):
            return self.entry.get_text()


    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default="", **kwargs): # modif C.L.
            gtk.FileChooserButton.__init__(self,
                                           _("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    # modif C.L.
    class ColorSelector(gimpui.ColorSelector):
        def __init__(self, default=gimpcolor.RGB(1.0, 0, 0), **kwargs):
            ret = gimpui.ColorSelector.__init__(self, default)
            self.connect('color-changed', lambda _: kwargs['dialg'].la_previ.invalidate())
            return ret
    class FontSelector(gimpui.FontSelector):
        def __init__(self, default="Sans", **kwargs):
            ret = gimpui.FontSelector.__init__(self, default)
            self.connect('font-set', lambda _1, _2, _3: kwargs['dialg'].la_previ.invalidate())
            return ret
    class PatternSelector(gimpui.PatternSelector):
        def __init__(self, default="", **kwargs):
            ret = gimpui.PatternSelector.__init__(self, default)
            self.connect('pattern-set', lambda *_: kwargs['dialg'].la_previ.invalidate())
            return ret
    class GradientSelector(gimpui.GradientSelector):
        def __init__(self, default="", **kwargs):
            ret = gimpui.GradientSelector.__init__(self, default)
            self.connect('gradient-set', lambda *_: kwargs['dialg'].la_previ.invalidate())
            return ret
    class PaletteSelector(gimpui.PaletteSelector):
        def __init__(self, default="", **kwargs):
            ret = gimpui.PaletteSelector.__init__(self, default)
            self.connect('palette-set', lambda *_: kwargs['dialg'].la_previ.invalidate())
            return ret
    # end modif C.L.
    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : ColorSelector, # modif C.L.
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : PatternSelector,  # modif C.L.
            PF_GRADIENT    : GradientSelector, # modif C.L.
            PF_PALETTE     : PaletteSelector,  # modif C.L.
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    # modif C.L.
    hboxPrincipal = gtk.HBox(spacing=10)
	
    #dialog.vbox.pack_start(hboxPrincipal)			        # modif MrQ
    swin = gtk.ScrolledWindow()                             # New line MrQ
    swin.add_with_viewport(hboxPrincipal)                   # New Line MrQ
    dialog.vbox.pack_start(swin)                            # New Line MrQ
    swin.show()                                             # New Line MrQ	

    hboxPrincipal.show()
    
    dialog.la_previ = gimpui.DrawablePreview(start_params[1])
    dialog.la_previ.set_update(False)						# disable Preview
    dialog.la_previ.set_size_request(512,512)               # New Line MrQ	
    dialog.la_previ.la_img = start_params[0]
    dialog.la_previ.la_layer = start_params[1]
    hboxPrincipal.pack_start(dialog.la_previ)

    dialog.set_default_size(1100, 680)  					# New Line MrQ
    dialog.la_previ.show()
    # end modif C.L.

    vbox = gtk.VBox(False, 2)
    vbox.set_border_width(2)
    #vbox.set_size_request(-10, 0)                           # modif C.L. + MrQ
    #~ dialog.vbox.pack_start(vbox)
    hboxPrincipal.pack_start(vbox) # modif C.L.
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)

    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})
    # modif C.L.
    def ifinvalidated(p_la_previ):
        
        # get dimension
        #===============
        x, y = p_la_previ.get_position()
        w, h = p_la_previ.get_size()
        
        # Effet
        #=========
        # has alpha
        lb_alpha = pdb.gimp_drawable_has_alpha(p_la_previ.la_layer)
        
        # img, layer temporaires
        tempImg = pdb.gimp_image_new(w, h, 0)
        #~ tempLayer = pdb.gimp_layer_new(tempImg, w, h, 0, 'temp', 100, 0)
        tempLayer = pdb.gimp_layer_new(tempImg, w, h, lb_alpha, 'temp', 100, 0)
        pdb.gimp_image_insert_layer(tempImg, tempLayer, None, 0)
        
        # copie orig dans tempo
        rgnPreviIn = p_la_previ.la_layer.get_pixel_rgn(x, y, w, h)
        rgnTemp = tempLayer.get_pixel_rgn(0, 0, w, h)
        rgnTemp[0:w, 0:h] = rgnPreviIn[x:x+w, y:y+h]
        
        # effect
        tempImg.undo_group_start()
        tempImg.disable_undo()
        
        params = []

        try:
            for wid in edit_wids:
                params.append(wid.get_value())
        except EntryValueError:
            warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
        else:
            try:
                paramsTot = (tempImg, tempLayer) + tuple(params)
                _set_defaults(proc_name, start_params + tuple(params))
                function(*paramsTot)
                
                res = tempImg.active_layer
                
                if not lb_alpha: pdb.gimp_layer_flatten(res)
                
                tempImg.undo_group_end()
                
                rgnRes = res.get_pixel_rgn(0, 0, w, h)
                
                # mise dans previ
                #================
                p_la_previ.draw_buffer(rgnRes[0:w, 0:h], int(3+lb_alpha)*w)
                
                pdb.gimp_image_delete(tempImg)
                
                #pdb.gimp_display_new(tempImg)
            except CancelError:
                pass
            except Exception as e:
                dialog.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                error_dialog(dialog, proc_name)
                raise
    # end modif C.L.
    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        dialog.la_previ.set_update(False)	# disable Preview
        gtk.main_quit()
    dialog.la_previ.connect('invalidated', ifinvalidated) # modif C.L.
    dialog.connect("response", response)
    iniX=0
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]

        tab, incr, rows, connect, cols = (0, 1, 1, None, 1)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE:
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW):
                wid = _edit_mapping[pf_type](def_val, params[i][4], dialg=dialog) # modif C.L.
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4],dialg=dialog)
                else:
                    wid = _edit_mapping[pf_type](def_val)				


            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val, dialg=dialog) # modif C.L.
            if pf_type!=PF_TITLE:
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', desc)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()
	
    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break

        min_args = i

    if len(script_params) > min_args:
        start_params = params[:min_args + 1]

        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE

    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])

    gimp.displays_flush()

    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message
	
	
#####################################################################################################################################################
############################################# END fu_preview 0.1.1 by Claude Lion $ J.F. Garcia #####################################################
#####################################################################################################################################################

#####################################################################################################################################################
#                                        START fill_rotate.py Created by Tin Tran http://gimplearn.net/
#####################################################################################################################################################

def python_tt_fill_rotate_diag(new_image, layer, angle, mirror):
	pdb.gimp_image_undo_group_start(new_image)
	pdb.gimp_context_push()
	pdb.gimp_selection_none(new_image)
	#some calculation and save dimensions
	old_width,old_height = layer.width,layer.height
	new_width,new_height = layer.width*3,layer.height*3
	layer_position = pdb.gimp_image_get_item_position(new_image,layer)
	#set canvas to be 3 times in height and width of layer.
	pdb.gimp_image_resize(new_image,new_width,new_height,old_width,old_height)

	layer_copy_L = pdb.gimp_layer_new_from_drawable(layer,new_image)
	pdb.gimp_image_insert_layer(new_image,layer_copy_L,None,layer_position)
	if (mirror == 1):
		item = pdb.gimp_item_transform_flip_simple(layer_copy_L,ORIENTATION_HORIZONTAL,True,0)
		if pdb.gimp_layer_is_floating_sel(item):
			pdb.gimp_floating_sel_anchor(item)
	pdb.gimp_layer_translate(layer_copy_L,-old_width,0)
	
	layer_copy_R = pdb.gimp_layer_new_from_drawable(layer_copy_L,new_image)
	pdb.gimp_image_insert_layer(new_image,layer_copy_R,None,layer_position)
	pdb.gimp_layer_translate(layer_copy_R,old_width*2,0)

	layer_copy_T = pdb.gimp_layer_new_from_drawable(layer,new_image)
	pdb.gimp_image_insert_layer(new_image,layer_copy_T,None,layer_position)
	if (mirror == 1):
		item = pdb.gimp_item_transform_flip_simple(layer_copy_T,ORIENTATION_VERTICAL,True,0)
		if pdb.gimp_layer_is_floating_sel(item):
			pdb.gimp_floating_sel_anchor(item)
	pdb.gimp_layer_translate(layer_copy_T,0,-old_height)

	layer_copy_B = pdb.gimp_layer_new_from_drawable(layer_copy_T,new_image)
	pdb.gimp_image_insert_layer(new_image,layer_copy_B,None,layer_position)
	pdb.gimp_layer_translate(layer_copy_B,0,old_height*2)

	layer_copy_TL = pdb.gimp_layer_new_from_drawable(layer,new_image)
	pdb.gimp_image_insert_layer(new_image,layer_copy_TL,None,layer_position)
	if (mirror == 1):
		item = pdb.gimp_item_transform_flip_simple(layer_copy_TL,ORIENTATION_VERTICAL,True,0)
		if pdb.gimp_layer_is_floating_sel(item):
			pdb.gimp_floating_sel_anchor(item)
		item = pdb.gimp_item_transform_flip_simple(layer_copy_TL,ORIENTATION_HORIZONTAL,True,0)
		if pdb.gimp_layer_is_floating_sel(item):
			pdb.gimp_floating_sel_anchor(item)
	pdb.gimp_layer_translate(layer_copy_TL,-old_width,-old_height)

	layer_copy_TR = pdb.gimp_layer_new_from_drawable(layer_copy_TL,new_image)
	pdb.gimp_image_insert_layer(new_image,layer_copy_TR,None,layer_position)
	pdb.gimp_layer_translate(layer_copy_TR,old_width*2,0)

	layer_copy_BL = pdb.gimp_layer_new_from_drawable(layer_copy_TL,new_image)
	pdb.gimp_image_insert_layer(new_image,layer_copy_BL,None,layer_position)
	pdb.gimp_layer_translate(layer_copy_BL,0,old_height*2)

	layer_copy_BR = pdb.gimp_layer_new_from_drawable(layer_copy_TL,new_image)
	pdb.gimp_image_insert_layer(new_image,layer_copy_BR,None,layer_position)
	pdb.gimp_layer_translate(layer_copy_BR,old_width*2,old_height*2)

	#merge down them all to one original layer.
	BR_merged = pdb.gimp_image_merge_down(new_image,layer_copy_BR,EXPAND_AS_NECESSARY)
	BL_merged = pdb.gimp_image_merge_down(new_image,BR_merged,EXPAND_AS_NECESSARY)
	TR_merged = pdb.gimp_image_merge_down(new_image,BL_merged,EXPAND_AS_NECESSARY)
	TL_merged = pdb.gimp_image_merge_down(new_image,TR_merged,EXPAND_AS_NECESSARY)
	B_merged = pdb.gimp_image_merge_down(new_image,TL_merged,EXPAND_AS_NECESSARY)
	T_merged = pdb.gimp_image_merge_down(new_image,B_merged,EXPAND_AS_NECESSARY)
	R_merged = pdb.gimp_image_merge_down(new_image,T_merged,EXPAND_AS_NECESSARY)
	L_merged = pdb.gimp_image_merge_down(new_image,R_merged,EXPAND_AS_NECESSARY)
	ALL_merged = L_merged

	angle = angle/180.0 * math.pi
	item = pdb.gimp_item_transform_rotate(ALL_merged,angle,True,0,0)
	if pdb.gimp_layer_is_floating_sel(item):
		pdb.gimp_floating_sel_anchor(item)

	#set to original dimensions
	pdb.gimp_image_resize(new_image,old_width,old_height,-old_width,-old_height)
	pdb.gimp_layer_resize_to_image_size(ALL_merged)
	pdb.gimp_item_set_name(ALL_merged,"New pattern")
	pdb.gimp_context_pop()
	pdb.gimp_image_undo_group_end(new_image)
	pdb.gimp_displays_flush()
#####################################################################################################################################################	
# 																END fill_rotate.py
#####################################################################################################################################################

####################################################################################################################
################# START gegl_command by Kevin Payne & Claude Lion ##################################################
####################################################################################################################

from sys import platform
from ctypes import *

def load_library (library_name):
	if platform == "linux" or platform == "linux2":
		library_name = library_name + '.so.0'
	elif platform == "win32":
		from ctypes.util import find_library
		library_name = find_library (library_name + "-0")
	else:
		raise BaseException ("TODO")
	return CDLL (library_name)

gimpC = load_library ('libgimp-2.0')


def gegl_batch(image, drawable, gegl_graph_string):
	
	class GeglBuffer(Structure):
		pass
	drawable_id = drawable.ID
	
	gegl = load_library ('libgegl-0.4')
	gegl.gegl_init (None, None)
	gimpC.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
	gimpC.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

	x = c_int (); y = c_int (); w = c_int (); h = c_int ()
	non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
	args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
	
	if non_empty:
		source = gimpC.gimp_drawable_get_buffer (drawable_id)
		target = gimpC.gimp_drawable_get_shadow_buffer (drawable_id)
		gegl.gegl_render_op (source, target, "gegl:gegl", *args)
		gegl.gegl_buffer_flush (target)
		gimpC.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
		gimpC.gimp_drawable_update (drawable_id, x, y, w, h)
		gimpC.gimp_displays_flush ()
		
####################################################################################################################
################# END gegl_command #################################################################################
####################################################################################################################

def centered (image,layer):
	wL = pdb.gimp_drawable_width(layer)
	hL = pdb.gimp_drawable_height(layer)
	wI = pdb.gimp_image_width(image)
	hI = pdb.gimp_image_height(image)
	pdb.gimp_image_set_active_layer(image,layer)

	sw=(wI-wL)/2
	sh=(hI-hL)/2		
	pdb.gimp_layer_set_offsets(layer, sw, sh)

def prepare_square_wxw(img, layer):	

	wl = pdb.gimp_drawable_width(layer)
	hl = pdb.gimp_drawable_height(layer)
	if wl%2==1:
		wx = wl -1
	else:
		wx = wl
	if hl%2==1:
		hx = hl -1
	else:
		hx = hl
	p=min(wx,hx)
	pdb.gimp_layer_resize(layer, p, p, 0, 0)
	#centered(img,layer)
	#pdb.gimp_layer_resize_to_image_size(layer)

def symetrize_layer (img,layer, mir, sel):

	pdb.gimp_selection_none(img)
	reg_layer = pdb.gimp_layer_new_from_drawable(layer, img)
	pdb.gimp_image_insert_layer(img, reg_layer, None, 0)
	W = pdb.gimp_drawable_width(reg_layer)
	H = pdb.gimp_drawable_height(reg_layer)		

	if mir == 1:	# "Left to Right"
		pdb.gimp_item_transform_flip_simple (reg_layer, 0, True, 0)
		pdb.gimp_image_select_polygon(img, 0, 10,(0,0, W/2,0, W/2,H, 0,H, 0,0))
		pdb.gimp_edit_clear (reg_layer)
		pdb.gimp_selection_invert(img)	
		pdb.gimp_edit_clear(layer)
		pdb.gimp_selection_none(img)			

	if mir == 2:	# "Right to Left"
		pdb.gimp_item_transform_flip_simple (reg_layer, 0, True, 0)
		pdb.gimp_image_select_polygon(img, 0, 10,(W/2,0, W,0, W,H, W/2,H, W/2,0))
		pdb.gimp_edit_clear (reg_layer)
		pdb.gimp_selection_invert(img)	
		pdb.gimp_edit_clear(layer)
		pdb.gimp_selection_none(img)			

	if mir == 3:	# "Top to Bottom"
		pdb.gimp_item_transform_flip_simple (reg_layer, 1, True, 0)
		pdb.gimp_image_select_polygon(img, 0, 10,(0,0, W,0, W,H/2, 0,H/2, 0,0))
		pdb.gimp_edit_clear (reg_layer)
		pdb.gimp_selection_invert(img)	
		pdb.gimp_edit_clear(layer)
		pdb.gimp_selection_none(img)			

	if mir == 4:	# "Bottom to Top"
		pdb.gimp_item_transform_flip_simple (reg_layer, 1, True, 0)
		pdb.gimp_image_select_polygon(img, 0, 10,(0,H/2, 0,H, W,H, W,H/2, 0,H/2))
		pdb.gimp_edit_clear (reg_layer)
		pdb.gimp_selection_invert(img)	
		pdb.gimp_edit_clear(layer)
		pdb.gimp_selection_none(img)			
		
	if mir != 0:
		m = pdb.gimp_image_merge_down(img, reg_layer, 0)

	
	if sel == 1:	# "Left Top"
		
		pdb.gimp_item_transform_flip_simple (reg_layer, 0, True, 1)
		pdb.gimp_item_transform_rotate_simple(reg_layer, ROTATE_90,True,0,0)
		pdb.gimp_image_select_polygon(img, 0, 8,(0,0, W,0, 0,H, 0,0))		
		pdb.gimp_edit_clear (reg_layer)
		pdb.gimp_selection_none(img)
		pdb.gimp_image_select_polygon(img, 0, 8,(W,0, W,H, 0,H, W,0))
		pdb.gimp_edit_clear(layer)
		pdb.gimp_selection_none(img)		
		pdb.gimp_image_merge_down(img, reg_layer, 0)# merge_type  EXPAND-AS-NECESSARY (0), CLIP-TO-IMAGE (1), CLIP-TO-BOTTOM-LAYER (2)

	if sel == 2:  # "Right Bottom"
		
		pdb.gimp_item_transform_flip_simple (reg_layer, 0, True, 1)
		pdb.gimp_item_transform_rotate_simple(reg_layer, ROTATE_90,True,0,0)	
		pdb.gimp_image_select_polygon(img, 0, 8,(W,0, W,H, 0,H, W,0))
		pdb.gimp_edit_clear (reg_layer)
		pdb.gimp_selection_none(img)
		pdb.gimp_image_select_polygon(img, 0, 8,(0,0, W,0, 0,H, 0,0))
		pdb.gimp_edit_clear(layer)
		pdb.gimp_selection_none(img)		
		pdb.gimp_image_merge_down(img, reg_layer, 0)# merge_type  EXPAND-AS-NECESSARY (0), CLIP-TO-IMAGE (1), CLIP-TO-BOTTOM-LAYER (2)

	if sel == 3:	# "Left Bottom"
		
		pdb.gimp_item_transform_flip_simple (reg_layer, 0, True, 0)
		pdb.gimp_item_transform_rotate_simple(reg_layer, ROTATE_270,True,0,0)
		pdb.gimp_image_select_polygon(img, 0, 8,(0,0, 0,H, W,H, 0,0))
		pdb.gimp_edit_clear (reg_layer)
		pdb.gimp_selection_none(img)
		pdb.gimp_image_select_polygon(img, 0, 8,(0,0, W,0, W,H, 0,0))
		pdb.gimp_edit_clear(layer)
		pdb.gimp_selection_none(img)		
		pdb.gimp_image_merge_down(img, reg_layer, 0)# merge_type  EXPAND-AS-NECESSARY (0), CLIP-TO-IMAGE (1), CLIP-TO-BOTTOM-LAYER (2)

	if sel == 4: # "Right Top"
	
		pdb.gimp_item_transform_flip_simple (reg_layer, 0, True, 0)
		pdb.gimp_item_transform_rotate_simple(reg_layer, ROTATE_270,True,0,0)
		pdb.gimp_image_select_polygon(img, 0, 8,(0,0, W,0, W,H, 0,0))
		pdb.gimp_edit_clear (reg_layer)
		pdb.gimp_selection_none(img)
		pdb.gimp_image_select_polygon(img, 0, 8,(0,0, 0,H, W,H, 0,0))
		pdb.gimp_edit_clear(layer)
		pdb.gimp_selection_none(img)		
		pdb.gimp_image_merge_down(img, reg_layer, 0)# merge_type  EXPAND-AS-NECESSARY (0), CLIP-TO-IMAGE (1), CLIP-TO-BOTTOM-LAYER (2)

	#pdb.gimp_image_remove_layer(img, copy)
####################################################################################################################
# palette_from_image.py
# creates palette from image.
# Created by Tin Tran 
# Download https://gimplearn.net/viewtopic.php?f=3&t=1637	
def palette_from_image_tt_pe(image, layer):
	hor_colors=16
	vert_colors=16
	pdb.gimp_image_undo_group_start(image)
	pdb.gimp_context_push()
	column_width = layer.width / hor_colors;
	row_height = layer.height / vert_colors;
	mid_offset_x = column_width / 2.0; #offset it so that we generally get the middle pixel.
	mid_offset_y = row_height / 2.0; #offset it so that we generally get the middle pixel.
	actual_name = pdb.gimp_palette_new("000-Palette")
	for iy in range(0,int(vert_colors)):
		y = iy * row_height + mid_offset_y;
		for ix in range(0,int(hor_colors)):
			x = ix * column_width + mid_offset_x;
			num_channels,pixel = pdb.gimp_drawable_get_pixel(layer,x,y)
			pdb.gimp_palette_add_entry(actual_name,"Color " + str(iy*16+ix+1),pixel[0:3])
	pdb.gimp_context_pop()
	pdb.gimp_image_undo_group_end(image)
	pdb.gimp_displays_flush()
	
########################################################################################################	
# Plugin  : palette-to-textile.py
# Author  : Arch.  
# Date    : 18 giugno 2009 
# Version : 1.0	
def palette_to_textile_pe(image, layer,height, width, palette):

    num_colors = pdb.gimp_palette_get_info(palette)
    num_segments = num_colors - 1
    gradient = pdb.gimp_gradient_new("ATGtemp")
    pdb.gimp_gradient_segment_range_split_uniform(gradient, 0, -1, num_segments)

    for color_number in range(0,num_segments):
        if (color_number == num_colors-1):color_number_next = 0
        else: color_number_next = color_number + 1
        color_left = pdb.gimp_palette_entry_get_color(palette, color_number)
        color_right = pdb.gimp_palette_entry_get_color(palette, color_number_next)
        pdb.gimp_gradient_segment_set_left_color(gradient, color_number, color_left, 100.0)
        pdb.gimp_gradient_segment_set_right_color(gradient, color_number, color_right, 100.0)
    pdb.gimp_context_set_gradient(gradient)
    layer = pdb.gimp_layer_new(image, width, height, RGBA_IMAGE, "ATG_0", 100, NORMAL_MODE)
    pdb.gimp_image_add_layer(image, layer, -1)
    pdb.gimp_edit_blend(layer, CUSTOM_MODE, NORMAL_MODE, GRADIENT_SEGMENT_LINEAR, 100, 0, REPEAT_NONE, FALSE, FALSE, 3, 1, TRUE, 0, 0, width, 0)
    layer2 = pdb.gimp_layer_new(image, width, height, RGBA_IMAGE, "ATG_1", 50, NORMAL_MODE)
    pdb.gimp_image_add_layer(image, layer2, -1)
    pdb.gimp_displays_flush()
    pdb.gimp_edit_blend(layer2, CUSTOM_MODE, NORMAL_MODE, GRADIENT_SEGMENT_LINEAR, 80, 0, REPEAT_NONE, FALSE, FALSE, 3, 1, TRUE, 0, 0, 0, height)
    pdb.gimp_gradient_delete(gradient)
    pattern_n=pdb.gimp_image_merge_down(image, pdb.gimp_image_get_layer_by_name(image, "ATG_0"), 0)
	
##########################################################################################################################################		

def Python_fu_Pattern_Editor(image, layer, select,typeScale,inScale,new_w,new_h,hor,ver,rot_180,angle,
	palet,name_palet, desaturate,colorize,color_c, opacity_color, add_efect,efect_opacity,efect_mode,brightness,contrast,
	square_sel, prep_square,use_mir, mir_1, mir_2, mir_3, mir_4, use_sel, sel_1, sel_2, sel_3, sel_4,
	array, a_iterations, a_xoffset, a_yoffset, a_mode, a_init, a_crop,
	kaleidoscope_g, m_angleh_g, r_angle_g, n_segs_g, kaleidoscope_s, iteration, angle_k, kaleidoscope_b, angular_tiles, smoothness,
	deform,amplitude,emboss,type,azimuth_g,elevation_g,depth_g,unsharp_mask, radius, amount, thresh, tile,
	fill_layer, code_info_name,name_pattern,save_type,save_as,code_info_save,
	help, ptab):

	# Store the GIMP's settings so they can be restored
	#gimp.context_push()
	# Set start undo point
	#pdb.gimp_image_undo_group_start(image)

	for layer in image.layers:
		pdb.gimp_image_set_active_layer(image,layer)
		pdb.gimp_item_set_visible(layer, 0)
	
	if select==0:
		temp_layer=pdb.gimp_layer_copy(layer,1)
		pdb.gimp_image_insert_layer(image, temp_layer, None, -1)
		pdb.gimp_image_raise_layer_to_top(image, temp_layer)
		temp_layer.name="New pattern"

	if select==1:
		width, height, bpp = pdb.gimp_pattern_get_info(pdb.gimp_context_get_pattern())
		temp_layer = pdb.gimp_layer_new(image,width,height, 1, "New pattern", 100, 0)	
		pdb.gimp_image_insert_layer(image, temp_layer, None, -1)
		pdb.gimp_drawable_fill(temp_layer, 4)
		#centered(image, temp_layer)
	pdb.gimp_item_set_visible(temp_layer, 1)	

	w = pdb.gimp_drawable_width(temp_layer)
	h = pdb.gimp_drawable_height(temp_layer)

	if palet:

		if palet==1:
			palette_from_image_tt_pe(image,temp_layer)
			palette_to_textile_pe(image, temp_layer,h, w, "000-Palette")
			pdb.gimp_palettes_refresh()
		if palet==2:
			palette_to_textile_pe(image, temp_layer,h, w, name_palet)			
		pattern_new=pdb.gimp_image_merge_down(image, pdb.gimp_image_get_layer_by_name(image, "ATG_1"), 0)
				
		temp_layer=pattern_new
		
	if hor:  # ("HORIZONTAL "u'\u21e8'""),   
		pdb.gimp_item_transform_flip_simple(temp_layer, 0, FALSE, w/2)
		
	if ver:	# _("VERTICAL "u'\u21e9'""),
		pdb.gimp_item_transform_flip_simple(temp_layer, 1, FALSE, h/2)

	if rot_180:	#  _("ROTATE 180"),
		pdb.gimp_item_transform_rotate_simple(temp_layer, 1, FALSE, w/2, h/2)

	if angle:
		python_tt_fill_rotate_diag(image, temp_layer, angle, 1)
		
	if select==0:
		centered(image, pdb.gimp_image_get_layer_by_name(image, "New pattern"))
####################################################################################################
	if desaturate == 1:		
		pdb.gimp_drawable_desaturate(pdb.gimp_image_get_layer_by_name(image, "New pattern"), 4)
		
	if colorize == 1:
		newlayer2 = pdb.gimp_layer_new_from_drawable(pdb.gimp_image_get_layer_by_name(image, "New pattern"), image)
		pdb.gimp_image_insert_layer(image, newlayer2, None, 0)
		pdb.gimp_context_set_foreground(color_c)
		pdb.gimp_drawable_fill(newlayer2, FILL_FOREGROUND)
		pdb.gimp_layer_set_mode(newlayer2, LAYER_MODE_SCREEN_LEGACY)
		pdb.gimp_layer_set_opacity(newlayer2, opacity_color)		
		efect_layer = pdb.gimp_image_merge_down(image, newlayer2, CLIP_TO_IMAGE)	# MrQ: leave the colorization layer to the user	

		#pdb.gimp_item_set_name(efect_layer,"New pattern")
		
	if add_efect!=18:
		pdb.gimp_selection_none(image)	
		efect_layer = pdb.gimp_layer_new_from_drawable(pdb.gimp_image_get_layer_by_name(image, "New pattern"), image)
		pdb.gimp_image_insert_layer(image, efect_layer, None, 0)
		pdb.gimp_item_set_name(efect_layer, "Efect layer")
		#pdb.gimp_image_select_item(image, 2, efect_layer)		

	if add_efect==1:	# Alien Map
		gegl_batch(image, efect_layer, "alien-map cpn-3-keep=1")
	if add_efect==2:	# Camouflage
		pdb.plug_in_gmic_qt(image,efect_layer, 1, 0, 'fx_camouflage 9,12,100,30,46,33,75,90,65,179,189,117,255,246,158')
	if add_efect==3:	# canvas			
		gegl_batch(image, efect_layer, 'texturize-canvas')
	if add_efect==4:	# crystal
		pdb.plug_in_gmic_qt(image,efect_layer, 1, 0, 'fx_crystal_background 10,25,0,100,1')			
	if add_efect==5:	# cubism			
		gegl_batch(image, efect_layer, 'cubism')
	if add_efect==6:	# Denim
		pdb.plug_in_gmic_qt(image,efect_layer, 1, 0,'Denim_samj 5,2,0,40,40,25,50,0,43,108,126,255')
	if add_efect==7:	# edge			
		gegl_batch(image, efect_layer, 'edge')
	if add_efect==8:	# edge-neon			
		gegl_batch(image, efect_layer, 'edge-neon')		
	if add_efect==9:	# edge-sobel		
		gegl_batch(image, efect_layer, 'edge-sobel')	
	if add_efect==10:	# engrave rgb-clip	
		gegl_batch(image, efect_layer, 'engrave rgb-clip')
	if add_efect==11:	# gradmap		
		pdb.plug_in_gradmap(image, efect_layer)
	if add_efect==12:	# mosaic
		gegl_batch(image, efect_layer, 'mosaic')
	if add_efect==13:	# pixelize
		gegl_batch(image, efect_layer, 'pixelize')		
	if add_efect==14:	# plasma
		gegl_batch(image, efect_layer, 'plasma')
	if add_efect==15:	# rainbow
		pdb.plug_in_gmic_qt(image,efect_layer, 1, 0, 'fx_rep_rainbowify 0,0,100,0,50,50')
	if add_efect==16:	# value-invert	
		gegl_batch(image, efect_layer, 'value-invert')	
	if add_efect==17:	# video-degradation	
		gegl_batch(image, efect_layer, 'video-degradation pattern=wide-striped additive=0')
	
	#if add_efect:
		#pdb.gimp_layer_resize_to_image_size(efect_layer)
	pdb.gimp_layer_set_opacity(efect_layer,efect_opacity)
	if efect_mode!=22 and efect_mode!=2:	
		pdb.gimp_layer_set_mode(efect_layer, efect_mode)
	efect_layer=pdb.gimp_image_merge_down(image, efect_layer, 0)
	pdb.gimp_selection_none(image)
	if palet==1:	
		pdb.gimp_palette_delete("000-Palette")
		#pdb.gimp_palette_delete("A New Palette #1")
################################################################################################################################		
		
	pdb.gimp_image_resize_to_layers(image)
    
	w = pdb.gimp_drawable_width(pdb.gimp_image_get_layer_by_name(image, "New pattern"))
	h = pdb.gimp_drawable_height(pdb.gimp_image_get_layer_by_name(image, "New pattern"))
	
	if prep_square:
		prepare_square_wxw(image, pdb.gimp_image_get_layer_by_name(image, "New pattern"))
		#centered(image,pdb.gimp_image_get_layer_by_name(image, "New pattern"))

	#if use_mir:		
	if mir_1:
		symetrize_layer (image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), 1, 0)
	if mir_2:		
		symetrize_layer (image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), 2, 0)
	if mir_3:
		symetrize_layer (image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), 3, 0)
	if mir_4:		
		symetrize_layer (image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), 4, 0)
			
	#if use_sel:			
	if sel_1:		
		symetrize_layer (image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), 0, 1)
	if sel_2:			
		symetrize_layer (image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), 0, 2)		
	if sel_3:		
		symetrize_layer (image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), 0, 3)
	if sel_4:			
		symetrize_layer (image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), 0, 4)

		
	if array:		
		pdb.plug_in_gmic_qt(image,pdb.gimp_image_get_layer_by_name(image, "New pattern"), 1, 0,"fx_array_mirror " + str(a_iterations) + "," + str(a_xoffset)+ "," + str(a_yoffset)+ "," + str(a_mode)+ "," + str(a_init)+ ",0," + str(a_crop))		
			
						
	if kaleidoscope_g:
		gegl_batch(image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), 'mirrors m-angleh=%d r-angle=%d n-segs=%d'  % (m_angleh_g, r_angle_g, n_segs_g))	
# gmic			

	if kaleidoscope_s:	# fx_symmetrizoscope 7,0,3,0	
		pdb.plug_in_gmic_qt(image,pdb.gimp_image_get_layer_by_name(image, "New pattern"), 1, 0,"fx_symmetrizoscope " + str(iteration) + "," + str(angle_k) +",3,0")
        
	if kaleidoscope_b:  #Blended fx_rotoidoscope 50,50,8,0.44,3	
		pdb.plug_in_gmic_qt(image,pdb.gimp_image_get_layer_by_name(image, "New pattern"), 1, 0,"fx_rotoidoscope 50,50," + str(angular_tiles) + "," + str(smoothness) + ",3")
	
	
	if typeScale==1:
		pdb.gimp_layer_scale(pdb.gimp_image_get_layer_by_name(image, "New pattern"), pdb.gimp_image_get_layer_by_name(image, "New pattern").width*inScale/100, pdb.gimp_image_get_layer_by_name(image, "New pattern").height*inScale/100, 1)
		
	if typeScale==2:
		pdb.gimp_layer_scale(pdb.gimp_image_get_layer_by_name(image, "New pattern"), new_w, new_h, 1)

	
	if deform:
		pdb.plug_in_gmic_qt(image,pdb.gimp_image_get_layer_by_name(image, "New pattern"), 1, 0, "deform " + str(amplitude))
		
	if emboss:
		if type==0:
			type_g='emboss'
		if type==1:
			type_g='bumpmap'
		gegl_batch(image, pdb.gimp_image_get_layer_by_name(image, "New pattern"),'emboss type=%s azimuth=%d elevation=%d depth=%d' % (type_g, azimuth_g, elevation_g, depth_g))	
	if unsharp_mask:
		gegl_batch(image, pdb.gimp_image_get_layer_by_name(image, "New pattern"),'unsharp-mask std-dev=%d scale=%d threshold=%d' % (radius, amount, thresh))
	if tile==1:
		gegl_batch(image, pdb.gimp_image_get_layer_by_name(image, "New pattern"),'tile-seamless')
		
	pdb.gimp_drawable_brightness_contrast(pdb.gimp_image_get_layer_by_name(image, "New pattern"), brightness, contrast)	

		
		
	if fill_layer==1:
		x=pdb.gimp_context_get_pattern()
	
		pdb.gimp_edit_copy(pdb.gimp_image_get_layer_by_name(image, "New pattern"))
		pdb.gimp_context_set_pattern(pdb.gimp_patterns_list('')[1][0])
		pdb.gimp_drawable_edit_fill(layer,FILL_PATTERN)
		image.clean_all()
		#pdb.gimp_image_raise_item_to_top(image, layer)
		pdb.gimp_item_set_visible(layer, True)
		pdb.gimp_context_set_pattern(x)
		pdb.gimp_item_set_visible(pdb.gimp_image_get_layer_by_name(image, "New pattern"), 0)
		

	if save_as:
		path=gimp.directory +  os.sep + 'patterns'
		if save_type==0:
			filename=path +  os.sep + name_pattern + ".pat"
			pdb.file_pat_save(image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), filename, filename, name_pattern)
		if save_type==1:
			filename=path +  os.sep + name_pattern + ".png"
			pdb.file_png_save(image, pdb.gimp_image_get_layer_by_name(image, "New pattern"), filename, filename, 0, 9, 0, 0, 0, 0, 0)		
		pdb.gimp_patterns_refresh()
		
	view = pdb.gimp_layer_new_from_visible(image, image, "Preview from Visible") 
	pdb.gimp_image_insert_layer(image, view, None, 0)		
	pdb.gimp_displays_flush()
	
	# Set end undo point
	#pdb.gimp_image_undo_group_end(image)

	# Restore the GIMP's settings
	#gimp.context_pop()	

hlp=" Resizing method 'Preview':\n 1. Disable preview.\n 2. Drag the corner.\n 3. Turn on the preview again."	\
+ "\n\n Only the size of the initial (active) layer is displayed, so the preview\n may show an image slightly different from the final result."	\
+ "\n\n Most transformations cause seamless patterns to no longer be \n seamless - so use gegl 'Tile Seamless' (in the tab: 'Deform')." \
+ "\n\n Sometimes it causes that the result is not very similar" \
+ "\n to the original pattern - then it is better to fill the layer" \
+ "\n with the selected pattern and manually edit (transform) the layer. " \
+ "\n\n The plugin works on the basis of a named layer: 'New pattern'.\n Don't use this name for layers before running the plug-in. " \
+ "\n\n Plugin preview works badly on selection - so You should use:\n 'Select' → 'None' before using plugin (!)." \
+ "\n\n If the initial layer used is larger than 512 px, the preview \n of the new pattern is not correct (because it is also scaled\n to the size of the preview). "	\
+ "\n\n Wish You have fun with this plugin:\n\n MareroQ."

register(
	"Edit_Active_Layer_or_Pattern",
	" Create new pattern from active layer or active pattern.Ver.9.\n Select parameters, then turn on 'Preview'.",
	"Create and save new pattern from active layer or active pattern.\nRequires an alpha channel",
	"MrQ",
	"GPL",
	"2022",
	"Pattern Editor... ",
	"RGBA",      # Alternately use RGB, RGB*, GRAY*, INDEXED etc.
	[

		(PF_IMAGE,    "image", "Input image", None),
		(PF_DRAWABLE, "layer", "Input layer", None),
		(PF_TOGGLE,   "select",   "Create from:", 0,[" Active Layer "," Active Pattern "],{'tab':0,'newLine':1}),
		(PF_OPTION,   "typeScale",   "To Scale:", 0,["Don't Use ","Percentage (%)","By size (in pixels)"],{'tab':0,'newLine':'0'}),
		(PF_SPINNER,  "inScale", "	Scale (%):", 100, (2, 500, 2),{'tab':0,'newLine':1}),
		(PF_SPINNER,  "new_w", "New Width (px):", 256, (2, 1000, 2),{'tab':0,'newLine':'0'}),
		(PF_SPINNER,  "new_h", "	New Height (px): ", 256, (2, 1000, 2),{'tab':0,'newLine':1}),		
		(PF_TOGGLE,   "hor",   "Flip HORIZONTAL:", 0,[("Don't Use "),(" YES")],{'tab':0,'newLine':'0'}),
		(PF_TOGGLE,   "ver",   "	Flip VERTICAL:", 0,[("Don't Use "),(" YES")],{'tab':0,'newLine':1}),
		(PF_TOGGLE,   "rot_180",   "ROTATE 180:", 0,[("Don't Use "),(" YES")],{'tab':0,'newLine':'0'}),
		(PF_SPINNER,  "angle", "	Rotate\n	(angle):", 0, (-360, 360, 1),{'tab':0,'newLine':1}),

		(PF_OPTION,  "palet",   "Checkered?", 0,["Don't Use "," From Colors Layer"," From Colors Palette"],{'tab':1,'newLine':'0'}),
		(PF_PALETTE, "name_palet", "  Palette:", "Bears",0,{'tab':1,'newLine':1}),		
		(PF_TOGGLE,  "desaturate",   "Desaturate?", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':1}),
		(PF_TOGGLE,  "colorize", "Colorize? ", 0 ,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':1}),		
		(PF_COLOR,   "color_c", "  color:", (230, 34, 34),0,{'newLine':'0','tab':1}),			
		(PF_SPINNER, "opacity_color", "  opacity:", 100, (1, 100, 1),{'newLine':'1','tab':1}),
		(PF_OPTION,  "add_efect",   "Add Effect?", 0,["Don't Use","Alien Map","Camouflage","Canvas","Crystal","Cubism","Denim","Edge","Edge Neon","Edge Sobel","Engrave","Gradient Map","Mosaic","Pixelize","Plasma","Rainbow","Value Invert","Video Degradation"],{'newLine':'0','tab':1}),		
		(PF_SPINNER, "efect_opacity", "  opacity: ", 85, (0, 100, 1),{'newLine':'1','tab':1}),
		(PF_OPTION, "mode", "Blend Mode: ", 22, 
		[" NORMAL-LEGACY (0) ",
		" DISSOLVE (1) ",
		" BEHIND-LEGACY\n(It will not be used) ",
		" MULTIPLY-LEGACY (3) ",
		" SCREEN-LEGACY (4) ",
		" OVERLAY-LEGACY (5) ",
		" DIFFERENCE-LEGACY (6) ",
		" ADDITION-LEGACY (7) ",
		" SUBTRACT-LEGACY (8) ",
		" DARKEN-ONLY-LEGACY (9) ",
		" LIGHTEN-ONLY-LEGACY (10) ",
		" HSV-HUE-LEGACY (11) ",
		" HSV-SATURATION-LEGACY (12) ",
		" HSL-COLOR-LEGACY (13) ",
		" HSV-VALUE-LEGACY (14) ",
		" DIVIDE-LEGACY (15) ",
		" DODGE-LEGACY (16) ",
		" BURN-LEGACY  (17) ",
		" HARDLIGHT-LEGACY (18) ",
		" SOFTLIGHT-LEGACY (19) ",
		" GRAIN-EXTRACT-LEGACY (20) ",
		" GRAIN-MERGE-LEGACY (21) ",
		" DON'T USE\n"u'\u2191'" Legacy  -  New (default) "u'\u2193'" ", 
		" OVERLAY (23) ",
		" LCH-HUE (24) ",
		" LCH-CHROMA (25) ",
		" LCH-COLOR (26) ",
		" LCH-LIGHTNESS (27) ", 
		" NORMAL (28) ",
		" BEHIND (29) ",
		" MULTIPLY (30) ",
		" SCREEN (31) ",
		" DIFFERENCE (32) ",
		" ADDITION (33) ",
		" SUBTRACT (34) ",
		" DARKEN-ONLY (35) ",
		" LIGHTEN-ONLY (36) ",
		" HSV-HUE (37) ",
		" HSV-SATURATION (38) ",
		" HSL-COLOR (39) ",
		" HSV-VALUE (40) ",
		" DIVIDE (41) ",
		" DODGE (42) ",
		" BURN (43) ",
		" HARDLIGHT (44) ",
		" SOFTLIGHT (45) ",
		" GRAIN-EXTRACT (46) ",
		" GRAIN-MERGE (47) ",
		" VIVID-LIGHT (48) ",
		" PIN-LIGHT (49) ",
		" LINEAR-LIGHT (50) ",
		" HARD-MIX (51) ",
		" EXCLUSION (52) ",
		" LINEAR-BURN (53) ",
		" LUMA-DARKEN-ONLY (54) ",
		" LUMA-LIGHTEN-ONLY (55) ",
		" LUMINANCE (56) ",
		],{'tab':1,}),		
		(PF_SPINNER, "brightness", _("Brightness:"), 0, (-0.5, 0.5, 0.01),{'newLine':'1','tab':1,}),
		(PF_SPINNER, "contrast", _("Contrast:"), 0, (-0.5, 0.5, 0.01),{'newLine':'1','tab':1}),
		
		(PF_TITLE,  "square_sel"  ,  "<b> * This options requires a square with an even number side length *\n\n                              PREPARE SQUARE</b>\n                    [crop to the short side= min (width,height)]:",(0.0,0.0), None,{'iniCol':0,'cols':6,'tab':2}),
		(PF_TOGGLE, "prep_square",   "", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':2}),	
		(PF_TITLE,  "use_mir"  ,  "<b>                           MIRROR BY CENTRAL AXIS:</b>\n                                            (choose an area)",(0.0,0.0), None,{'iniCol':0,'cols':6,'tab':2}),		
		(PF_TOGGLE, "mir_1",   u'  \u21E8'" Left to Right ", 0,[("Don't Use "),(" YES")],{'tab':2,'newLine':'0'}),
		(PF_TOGGLE, "mir_2",   u'  \u21E6'" Right to Left ", 0,[("Don't Use "),(" YES")],{'tab':2}),
		(PF_TOGGLE, "mir_3",   u'  \u21E9'" Top to Bottom ", 0,[("Don't Use "),(" YES")],{'tab':2,'newLine':'0'}),
		(PF_TOGGLE, "mir_4",   u'  \u21E7'" Bottom to Top ", 0,[("Don't Use "),(" YES")],{'tab':2}),		
		(PF_TITLE,  "use_sel"  ,  "<b>                           SYMMETRY BY DIAGONAL:</b>\n                                            (choose an area).",(0.0,0.0), None,{'iniCol':0,'cols':6,'tab':2}),		
		(PF_TOGGLE, "sel_1",   u'  \u25e4'" Left Top ", 0,[("Don't Use "),(" YES")],{'tab':2,'newLine':'0'}),
		(PF_TOGGLE, "sel_2",   u'  \u25e2'" Right Bottom ", 0,[("Don't Use "),(" YES")],{'tab':2}),
		(PF_TOGGLE, "sel_3",   u'  \u25e3'" Left Bottom ", 0,[("Don't Use "),(" YES")],{'tab':2,'newLine':'0'}),		
		(PF_TOGGLE, "sel_4",   u'  \u25e5'" Right Top ", 0,[("Don't Use "),(" YES")],{'tab':2}),		

		(PF_TOGGLE,  "array",   "G'MIC Array\n[Mirrored]:", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':3}),
		(PF_SPINNER, "a_iterations", " - Iterations: ", 1, (1, 10 , 1),{'tab':3}),
		(PF_SPINNER, "a_xoffset", " - X-Offset (%): ", 0, (0, 100 , 1),{'tab':3}),
		(PF_SPINNER, "a_yoffset", " - Y-Offset (%): ", 0, (0, 100 , 1),{'tab':3}),
		(PF_OPTION,  "a_mode",   " - Array Mode: ", 2,["X-Axis","Y-Axis","XY-Axes","2XY-Axes"],{'tab':3}),	
		(PF_OPTION,  "a_init",   " - Initialization:", 0,[" Original "," Mirror X "," Mirror Y "," Rotate 90 Deg. "," Rotate 180 Deg. "," Rotate 270 Deg. "],{'tab':3}),			
		(PF_SPINNER, "a_crop", " - Crop (%): ", 0, (0, 100 , 1),{'tab':3}),	
		
		(PF_TOGGLE,  "kaleidoscope_g",   "GEGL Kaleidoscopic :  ", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':4}),
		(PF_SPINNER, "m_angleh_g", " - Mirror rotation: ", 0, (0, 180 , 1),{'tab':4}),
		(PF_SPINNER, "r_angle_g", " - Result rotation: ", 4, (0, 360 , 1),{'tab':4}),
		(PF_SPINNER, "n_segs_g", " - Mirrors: ", 4, (2, 24 , 1),{'tab':4}),		
		(PF_TOGGLE,  "kaleidoscope_s",   "G'MIC Kaleidoscope \n(Symmetry):  ", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':4}),
		(PF_SPINNER, "iteration", " - iteration: ", 4, (1, 32 , 1),{'tab':4}),
		(PF_SPINNER, "angle_k", " - angle: ", 0, (0, 360 , 1),{'tab':4}),        
		(PF_TOGGLE,  "kaleidoscope_b",   "G'MIC Kaleidoscope \n(Blended):  ", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':4}),
		(PF_SPINNER, "angular_tiles", " - angular tiles: ", 10, (1, 72 , 1),{'tab':4}),
		(PF_SPINNER, "smoothness", " - smoothness: ", 0.5, (0, 5 , 0.1),{'tab':4}), 

		(PF_TOGGLE,  "deform",   "Use Randomly\nDeform:", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':5}),
		(PF_SPINNER, "amplitude", " - amplitude:", 10, (0, 100 , 1),{'tab':5}),		
		(PF_TOGGLE,  "emboss",   "Use Emboss:", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':5}),
		(PF_TOGGLE,  "type",   " - Type:", 1,[" Emboss "," Bumpmap "],{'tab':5}),		
		(PF_SPINNER, "azimuth_g", " - Azimuth:", 30, (0, 360, 1),{'tab':5}),
		(PF_SPINNER, "elevation_g", " - Elevation:", 45, (0, 180, 1),{'tab':5}),    
		(PF_SPINNER, "depth_g", " - Depth:", 20, (1, 100, 1),{'tab':5}),	
		(PF_TOGGLE,  "unsharp-mask",   "Use Unsharp-mask:", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':5}),			
		(PF_SPINNER, "radius", " - Radius:", 10.0, (0, 1500, 0.1),{'tab':5}),		
		(PF_SPINNER, "amount", " - Amount:", 5.0, (0, 10, 0.1),{'tab':5}),		
		(PF_SPINNER, "thresh", " - Threshold:", 0.5, (0, 1, 0.01),{'tab':5}),
		(PF_TOGGLE,  "tile",   "Use Tile Seamless:", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':5}),
	
		(PF_TOGGLE, "fill_layer",   "Fill active layer:", 0,["Function inactive\n  *Click to Start*","Function Active\n *Click to Stop*"],{'tab':6}),
		(PF_TITLE,  "code_info_name"  , "<b> 	Pattern is saved in the gimp home directory: .../patterns/.\n		  Don't forget to enter a unique pattern name \n		(to avoid being overwritten - without warning).</b>",(0.0,0.0), None  ,{'iniCol':0,'cols':6,'tab':6}),			
		(PF_STRING, "name_pattern",  ' Name Pattern:', "Your name",1,{'tab':6}),
		(PF_TOGGLE, "save_type",   " File type\n (extension): ", 0,[(" .pat "),(" .png")],{'tab':6}),        
		(PF_TOGGLE, "save_as",   "Save new Pattern: ", 0,[("Don't Use - Click to Save. "),(" YES")],{'tab':6}),
		(PF_TITLE,  "code_info_save"  , "<b> 	       After clicking 'Save' the file will be saved\n		        and the patterns refreshed.\n		   Don't forget to switch off after saving.</b>",(0.0,0.0), None  ,{'iniCol':0,'cols':6,'tab':6}),			
		
		(PF_TITLE, "help"  , hlp ,(10,10), None           ,{'tab':7}),		
		(PF_TAB,   "ptab"    , "tabs" , 0, (' Transform'," Effect"," Symmetrize"," Array"," Kaleidoscope",' Deform'," Fill & Save" , "Help"),{'newLine':'1','tab':7} ),		
	],
	[],
	Python_fu_Pattern_Editor, menu="<Image>/Edit/")

main()